@@ -6,7 +6,7 @@ import warnings
 import numpy as np
 from scipy import fft, signal
 from IPython.display import display
-from miniaudio import get_file_info, decode_file, SampleFormat, convert_frames
+from miniaudio import get_file_info, decode_file, SampleFormat, DecodeError
 from bokeh.plotting import figure, show
 from bokeh.layouts import gridplot
 from bokeh.models import ColumnDataSource
@@ -15,7 +15,7 @@ from bokeh.models.ranges import DataRange1d
 from bokeh.models.tools import PanTool, BoxZoomTool, WheelZoomTool, ZoomInTool, ZoomOutTool, SaveTool, ResetTool, HoverTool, InspectTool
 from bokeh.palettes import Viridis256
 from bokeh.io import output_notebook
-from .util import Audio
+from .util import Audio, _read_aiff_file, _resample
 from .tf_rep import mel_filterbank, twelve_tet_filterbank, _fft_size_for_spacing
 
 
@@ -95,19 +95,24 @@ def same_type_as(output_samples, input_signal):
 
 
 class Signal(ABC):
-
+    """Abstract base class defining a generic signal.
+    """
     @abstractmethod
     def plot(self, **fig_args):
         pass
 
+
     def _repr_html_(self):
         return show(self.plot())
 
+
     def display(self, **fig_args):
         show(self.plot(**fig_args))
 
-class TimeSignal(Signal):
 
+class TimeSignal(Signal):
+    """A class representing a sampled time-signal.
+    """
     def __init__(self, in_data, samplerate=None):
         if isinstance(in_data, Spectrogram):
             if in_data.spacing != 'linear':
@@ -116,13 +121,16 @@ class TimeSignal(Signal):
                 raise ValueError('A spectrogram created with this combination of parameters cannot be inverted')
             if not signal.check_COLA(in_data.window, in_data._frame_size, in_data._overlap_size):
                 warnings.warn('A spectrogram created with this combination of parameters cannot be perfectly inverted')
-            self.times, self.samples = signal.istft(in_data.complex_spectrogram, in_data._samplerate, in_data.window, in_data._frame_size, in_data._overlap_size, in_data._fft_size)
+            self.times, self.samples = signal.istft(
+                in_data.complex, in_data._input_samplerate, in_data.window,
+                in_data._frame_size, in_data._overlap_size, in_data._fft_size,
+            )
             self.samples = np.clip(self.samples, -1, 1)
-            self.samplerate = in_data._samplerate
+            self.samplerate = in_data._input_samplerate
         else:
             if isinstance(in_data, Spectrum):
-                self.samples = fft.irfft(in_data.complex_spectrum, in_data._fft_size, norm='forward')
-                self.samplerate = in_data._samplerate
+                self.samples = fft.irfft(in_data.complex, in_data._fft_size, norm='forward')
+                self.samplerate = in_data._input_samplerate
             else:
                 if samplerate is None:
                     raise ValueError('Specify sample rate when creating TimeSignal from samples')
@@ -130,13 +138,14 @@ class TimeSignal(Signal):
                 self.samplerate = samplerate
             self.times = np.arange(len(self.samples)) / self.samplerate
 
+
     def plot(self, **fig_args):
         default_args = {
             'width': 900, 'height': 300,
             'x_axis_label': 'time [s]', 'y_axis_label': 'amplitude',
             'tools': 'pan,wheel_zoom,box_zoom,zoom_in,zoom_out,save,reset',
             'active_drag': 'pan',
-            'output_backend': 'webgl',
+            'toolbar_location': 'above',
         }
         fig = figure(**{**default_args, **fig_args})
         fig.line(self.times, self.samples, line_width=2)
@@ -194,31 +203,60 @@ class TimeSignal(Signal):
         resampled_signal = deepcopy(self)
         if samplerate != self.samplerate:
             resampled_signal.samplerate = samplerate
-            resampled_signal.samples = np.frombuffer(convert_frames(
-                SampleFormat.FLOAT32, 1, self.samplerate, self.samples.astype(np.float32).tobytes(),
-                SampleFormat.FLOAT32, 1, samplerate
-            ), dtype=np.float32)
+            resampled_signal.samples = _resample(self.samples, self.samplerate, samplerate)
             resampled_signal.times = np.arange(len(resampled_signal.samples)) / samplerate
-            resampled_signal.samples = np.clip(resampled_signal.samples, -1, 1)
         return resampled_signal
 
 
-class AudioSignal(TimeSignal):
+    def frames(self, frame_duration=None, frame_size=None, step_duration=None, step_size=None, step_ratio=None):
+        return [same_type_as(frame_samples, self) for frame_samples in self.frame_data(
+            frame_duration, frame_size, step_duration, step_size, step_ratio).T]
 
+
+    def frame_data(self, frame_duration=None, frame_size=None, step_duration=None, step_size=None, step_ratio=None):
+        frame_size, overlap_size = _parse_frame_params(frame_duration, frame_size, step_duration, step_size, step_ratio,
+                                                       self.samplerate, 'linear', None)
+        step_size = frame_size - overlap_size
+        samples = np.pad(self.samples, (frame_size//2, 0), 'constant')
+        num_frames = ((len(samples)-frame_size//2) // step_size) + 1
+        num_fill_samples = (num_frames-1)*step_size+frame_size-len(samples)
+        if num_fill_samples > 0:
+            samples = np.pad(samples, (0, num_fill_samples), 'constant')
+        return np.array([samples[i:i+frame_size] for i in range(0, len(samples)-frame_size+1, step_size)]).T
+
+
+class AudioSignal(TimeSignal):
+    """A class representing an audio signal with sample values between -1 and 1.
+    
+    Has playback functionality through the play() method.
+    """
     def __init__(self, in_data, samplerate=None):
         if isinstance(in_data, (str, Path)):
-            if samplerate is None:
-                file_info = get_file_info(in_data)
-                samplerate = file_info.sample_rate
-            decoded_file = decode_file(filename=in_data, output_format=SampleFormat.FLOAT32, nchannels=1, sample_rate=samplerate)
-            samples = np.asarray(decoded_file.samples)
+            try:
+                if samplerate is None:
+                    file_info = get_file_info(in_data)
+                    samplerate = file_info.sample_rate
+                decoded_file = decode_file(filename=in_data, output_format=SampleFormat.FLOAT32,
+                                           nchannels=1, sample_rate=samplerate)
+                samples = np.asarray(decoded_file.samples)
+            except DecodeError as decode_error:
+                try:
+                    samples, file_samplerate = _read_aiff_file(in_data, mix_to_mono=True)
+                    if samplerate is None:
+                        samplerate = file_samplerate
+                    elif samplerate != file_samplerate:
+                        samples = _resample(samples, file_samplerate, samplerate)
+                except:
+                    raise decode_error
             super().__init__(samples, samplerate)
         else:
             super().__init__(in_data, samplerate)
 
+
     def play(self, normalize=False):
         return display(Audio(self.samples, rate=self.samplerate, normalize=normalize))
 
+
     def plot(self, **fig_args):
         default_args = {
             'width': 900, 'height': 300,
@@ -248,7 +286,8 @@ class AudioSignal(TimeSignal):
 
 
 class Spectrum(Signal):
-
+    """ A class representing a complex spectrum.
+    """
     def __init__(
         self, in_data, *,
         dB=True, min_freq=None, max_freq=None, num_bins=None,
@@ -257,7 +296,7 @@ class Spectrum(Signal):
     ):
         samples, samplerate = get_samples_and_rate(in_data, samplerate)
 
-        self._samplerate = samplerate
+        self._input_samplerate = samplerate
         self.spacing = spacing
         self.window = window
         self.exponent = exponent
@@ -382,20 +421,22 @@ class Spectrum(Signal):
 
 
     @property
-    def complex_spectrum(self):
+    def complex(self):
         if self.dB:
             magnitude = 10 ** (self.magnitude / 20)
         else:
             magnitude = self.magnitude ** (1/self.exponent)
         if self._norm_single_side_band:
             if self._fft_size % 2 == 0:
-                self.magnitude[1:-1] /= sqrt(2)
+                magnitude[1:-1] /= sqrt(2)
             else:
-                self.magnitude[1:] /= sqrt(2)
+                magnitude[1:] /= sqrt(2)
         return magnitude * np.exp(1j*self.phase)
 
 
 class PowerSpectrum(Spectrum):
+    """A class representing a complex power spectrum.
+    """
     def __init__(self, in_data, **kwargs):
         exponent = kwargs.pop('exponent', None)
         if exponent is not None and exponent != 2:
@@ -404,44 +445,27 @@ class PowerSpectrum(Spectrum):
 
 
 class Spectrogram(Signal):
-
+    """A class representing a complex spectrogram.
+    """
     def __init__(
         self, input_signal, *,
         frame_duration=None, frame_size=None, step_duration=None, step_size=None, step_ratio=None,
-        dB=True, min_freq=None, max_freq=None, num_bins=None,
+        dB=True, min_freq=None, max_freq=None, num_bins=None, norm_single_side_band=False,
         spacing='linear', window='hann', exponent=1, samplerate=None,
     ):
         samples, samplerate = get_samples_and_rate(input_signal, samplerate)
 
         self._num_bins = num_bins
+        self._norm_single_side_band = norm_single_side_band
         self.spacing = spacing
         self.window = window
         self.exponent = exponent
         self.dB = dB
+        self._input_samplerate = samplerate
 
-        self._samplerate = samplerate
-        frame_specs = sum([frame_duration is not None, frame_size is not None])
-        if frame_specs != 1:
-            if spacing == 'linear':
-                raise ValueError('Specify a frame width either as duration in seconds or size in samples')
-            elif min_freq is None:
-                raise ValueError('Specify either a minimum frequency or a frame width in seconds or in samples')
-        if frame_duration is not None:
-            self._frame_size = round(frame_duration * samplerate)
-        elif frame_size is not None:
-            self._frame_size = frame_size
-        elif min_freq is not None and spacing != 'linear':
-            self._frame_size = ceil(samplerate / min_freq)
-        if sum([step_duration is not None, step_size is not None, step_ratio is not None]) > 1:
-            raise ValueError('Specify a frame step either as duration in seconds, size in samples or ratio')
-        if step_duration is not None:
-            self._overlap_size = round((frame_duration-step_duration) * samplerate)
-        elif step_size is not None:
-            self._overlap_size = self._frame_size - step_size
-        elif step_ratio is not None:
-            self._overlap_size = round((1-step_ratio) * self._frame_size)
-        else:
-            self._overlap_size = self._frame_size // 2
+        self._frame_size, self._overlap_size = _parse_frame_params(frame_duration, frame_size,
+            step_duration,step_size, step_ratio, samplerate, spacing, min_freq)
+        self.samplerate = samplerate / (self._frame_size - self._overlap_size)
 
         if max_freq is None:
             max_freq = samplerate / 2
@@ -459,11 +483,18 @@ class Spectrogram(Signal):
             raise ValueError('The frequency spacing needs to be one of "linear", "log" or "mel"')
 
         self._fft_size = max(self._frame_size, _fft_size_for_spacing(spacing, num_bins, samplerate, min_freq, max_freq))
-        _, self.times, spectrogram = signal.stft(samples, fs=samplerate, window=window, nperseg=self._frame_size, noverlap=self._overlap_size, nfft=self._fft_size)
-        filtered_spectrogram, self.frequencies, self._bin_values = _spacing_filter(spectrogram, spacing, samplerate, min_freq, max_freq, self._fft_size, num_bins)
+        _, self.times, spectrogram = signal.stft(samples, fs=samplerate, window=window, nperseg=self._frame_size,
+                                                 noverlap=self._overlap_size, nfft=self._fft_size, padded=False)
+        filtered_spectrogram, self.frequencies, self._bin_values = _spacing_filter(
+            spectrogram, spacing, samplerate, min_freq, max_freq, self._fft_size, num_bins)
         self.magnitude = np.abs(filtered_spectrogram)
         self.phase = np.angle(filtered_spectrogram)
 
+        if norm_single_side_band:
+            if self._fft_size % 2 == 0:
+                self.magnitude[1:-1] *= sqrt(2)
+            else:
+                self.magnitude[1:] *= sqrt(2)
         if dB:
             self.magnitude = 20 * np.log10(np.clip(self.magnitude, 1e-6, None))
         else:
@@ -506,7 +537,7 @@ class Spectrogram(Signal):
             fig.x_range.range_padding = 0
         if isinstance(fig.y_range, DataRange1d):
             fig.y_range.range_padding = 0
-        step_time = (self._frame_size - self._overlap_size) / self._samplerate
+        step_time = 1 / self.samplerate
         if [t for t in fig.tools if isinstance(t, InspectTool)]:
             all_times = np.broadcast_to(self.times, self.magnitude.shape)
             all_freqs = np.broadcast_to(self.frequencies.reshape(-1, 1), self.magnitude.shape)
@@ -522,12 +553,39 @@ class Spectrogram(Signal):
         return fig
 
 
+    def __len__(self):
+        return len(self.times)
+
+
+    def rms(self):
+        return sqrt(self.power(dB=False))
+
+
+    def power(self, dB=True):
+        if self.dB:
+            power_per_freq = 10 ** (self.magnitude / 10)
+        else:
+            power_per_freq = self.magnitude ** (2/self.exponent)
+        if self._norm_single_side_band:
+            power = np.sum(power_per_freq, axis=0)
+        elif self._fft_size % 2 == 0:
+            power = power_per_freq[0] + 2*np.sum(power_per_freq[1:-1], axis=0) + power_per_freq[-1]
+        else:
+            power = power_per_freq[0] + 2*np.sum(power_per_freq[1:], axis=0)
+        return 10*np.log10(power) if dB else power
+
+
     @property
-    def complex_spectrogram(self):
+    def complex(self):
         if self.dB:
             magnitude = 10 ** (self.magnitude / 20)
         else:
             magnitude = self.magnitude ** (1/self.exponent)
+        if self._norm_single_side_band:
+            if self._fft_size % 2 == 0:
+                magnitude[1:-1] /= sqrt(2)
+            else:
+                magnitude[1:] /= sqrt(2)
         return magnitude * np.exp(1j*self.phase)
 
 
@@ -541,8 +599,8 @@ class Spectrogram(Signal):
         else:
             min_freq, max_freq = self.frequencies[0], self.frequencies[-1]
         spectrum = Spectrum(np.zeros(self._frame_size), dB=self.dB, min_freq=min_freq, max_freq=max_freq, 
-            num_bins=self._num_bins, norm_single_side_band=False, spacing=self.spacing,
-            window=self.window, exponent=self.exponent, samplerate=self._samplerate)
+            num_bins=self._num_bins, norm_single_side_band=self._norm_single_side_band, spacing=self.spacing,
+            window=self.window, exponent=self.exponent, samplerate=self._input_samplerate)
         spectrum._fft_size = self._fft_size
         spectrum.magnitude = self.magnitude[:, index]
         spectrum.phase = self.phase[:, index]
@@ -550,6 +608,33 @@ class Spectrogram(Signal):
         return spectrum
 
 
+def _parse_frame_params(
+    frame_duration, frame_size, step_duration, step_size, step_ratio, samplerate, spacing, min_freq
+):
+    frame_specs = sum([frame_duration is not None, frame_size is not None])
+    if frame_specs != 1:
+        if spacing == 'linear':
+            raise ValueError('Specify a frame width either as duration in seconds or size in samples')
+        if min_freq is None:
+            raise ValueError('Specify either a minimum frequency or a frame width in seconds or in samples')
+    if frame_duration is not None:
+        frame_size = round(frame_duration * samplerate)
+    elif frame_size is None and min_freq is not None and spacing != 'linear':
+        frame_size = ceil(samplerate / min_freq)
+
+    if sum([step_duration is not None, step_size is not None, step_ratio is not None]) > 1:
+        raise ValueError('Specify a frame step either as duration in seconds, size in samples or ratio')
+    if step_duration is not None:
+        overlap_size = round((frame_duration-step_duration) * samplerate)
+    elif step_size is not None:
+        overlap_size = frame_size - step_size
+    elif step_ratio is not None:
+        overlap_size = round((1-step_ratio) * frame_size)
+    else:
+        overlap_size = frame_size // 2
+    return frame_size, overlap_size
+
+
 def _spacing_filter(spectral_values, spacing, samplerate, min_freq, max_freq, fft_size, num_bins):
     if spacing == 'linear':
         lin_frequencies = np.arange(len(spectral_values)) * samplerate / fft_size
